@@ -2,7 +2,6 @@
 //!
 //! The crate only provides a summary of the parameters.
 //! For more detailed documentation, see manpage.
-#![warn(rust_2018_idioms, unused_qualifications)]
 
 #[macro_use]
 mod util;
@@ -66,7 +65,11 @@ where
 }
 
 /// The parameters that were used to construct an [`IoUring`].
+///
+/// This type is a transparent wrapper over the system structure `io_uring_params`. A value can be
+/// (unsafely) created from any properly laid-out and initialized memory representation.
 #[derive(Clone)]
+#[repr(transparent)]
 pub struct Parameters(sys::io_uring_params);
 
 unsafe impl<S: squeue::EntryMarker, C: cqueue::EntryMarker> Send for IoUring<S, C> {}
@@ -81,6 +84,17 @@ impl IoUring<squeue::Entry, cqueue::Entry> {
     pub fn new(entries: u32) -> io::Result<Self> {
         Self::builder().build(entries)
     }
+
+    /// Create an `IoUring` instance from a pre-opened file descriptor.
+    ///
+    /// # Safety
+    ///
+    /// The caller must uphold that the file descriptor is owned and refers to a uring. The
+    /// `params` argument must be equivalent to the those previously filled in by the kernel when
+    /// the provided ring was created.
+    pub unsafe fn from_fd(fd: RawFd, params: Parameters) -> io::Result<Self> {
+        Self::with_fd_and_params(OwnedFd::from_raw_fd(fd), params.0)
+    }
 }
 
 impl<S: squeue::EntryMarker, C: cqueue::EntryMarker> IoUring<S, C> {
@@ -103,6 +117,11 @@ impl<S: squeue::EntryMarker, C: cqueue::EntryMarker> IoUring<S, C> {
     }
 
     fn with_params(entries: u32, mut p: sys::io_uring_params) -> io::Result<Self> {
+        let fd: OwnedFd = unsafe { OwnedFd::from_raw_fd(sys::io_uring_setup(entries, &mut p)?) };
+        unsafe { Self::with_fd_and_params(fd, p) }
+    }
+
+    unsafe fn with_fd_and_params(fd: OwnedFd, p: sys::io_uring_params) -> io::Result<Self> {
         // NOTE: The `SubmissionQueue` and `CompletionQueue` are references,
         // and their lifetime can never exceed `MemoryMap`.
         //
@@ -149,8 +168,6 @@ impl<S: squeue::EntryMarker, C: cqueue::EntryMarker> IoUring<S, C> {
             }
         }
 
-        let fd: OwnedFd = unsafe { OwnedFd::from_raw_fd(sys::io_uring_setup(entries, &mut p)?) };
-
         let (mm, sq, cq) = unsafe { setup_queue(&fd, &p)? };
 
         Ok(IoUring {
@@ -337,7 +354,6 @@ impl<S: squeue::EntryMarker, C: cqueue::EntryMarker> Builder<S, C> {
     /// events. You are only able to [register restrictions](Submitter::register_restrictions) when
     /// the rings are disabled due to concurrency issues. You can enable the rings with
     /// [`Submitter::register_enable_rings`]. Available since 5.10.
-
     pub fn setup_r_disabled(&mut self) -> &mut Self {
         self.params.flags |= sys::IORING_SETUP_R_DISABLED;
         self
@@ -573,11 +589,24 @@ impl Parameters {
         self.0.features & sys::IORING_FEAT_LINKED_FILE != 0
     }
 
+    /// Whether the kernel supports `IORING_RECVSEND_BUNDLE`.
+    ///
+    /// This feature allows sending and recieving multiple buffers as a single bundle. Available
+    /// since kernel 6.10.
+    pub fn is_feature_recvsend_bundle(&self) -> bool {
+        self.0.features & sys::IORING_FEAT_RECVSEND_BUNDLE != 0
+    }
+
     /// The number of submission queue entries allocated.
     pub fn sq_entries(&self) -> u32 {
         self.0.sq_entries
     }
 
+    /// The idle time of the SQ poll thread in milliseconds.
+    pub fn sq_thread_idle(&self) -> u32 {
+        self.0.sq_thread_idle
+    }
+
     /// The number of completion queue entries allocated.
     pub fn cq_entries(&self) -> u32 {
         self.0.cq_entries
