@@ -59,14 +59,14 @@ impl<'a> Submitter<'a> {
     #[inline]
     fn sq_need_wakeup(&self) -> bool {
         unsafe {
-            (*self.sq_flags).load(atomic::Ordering::Acquire) & sys::IORING_SQ_NEED_WAKEUP != 0
+            (*self.sq_flags).load(atomic::Ordering::Relaxed) & sys::IORING_SQ_NEED_WAKEUP != 0
         }
     }
 
     /// CQ ring is overflown
     fn sq_cq_overflow(&self) -> bool {
         unsafe {
-            (*self.sq_flags).load(atomic::Ordering::Acquire) & sys::IORING_SQ_CQ_OVERFLOW != 0
+            (*self.sq_flags).load(atomic::Ordering::Relaxed) & sys::IORING_SQ_CQ_OVERFLOW != 0
         }
     }
 
@@ -119,17 +119,36 @@ impl<'a> Submitter<'a> {
         // each cause an atomic load of the same variable, self.sq_flags.
         // In the hottest paths, when a server is running with sqpoll,
         // this is going to be hit twice, when once would be sufficient.
+        // However, consider that the `SeqCst` barrier required for interpreting
+        // the IORING_ENTER_SQ_WAKEUP bit is required in all paths where sqpoll
+        // is setup when consolidating the reads.
 
-        if want > 0 || self.params.is_setup_iopoll() || self.sq_cq_overflow() {
+        let sq_cq_overflow = self.sq_cq_overflow();
+
+        // When IORING_FEAT_NODROP is enabled and CQ overflows, the kernel buffers
+        // completion events internally but doesn't automatically flush them when
+        // CQ space becomes available. We must explicitly call io_uring_enter()
+        // to flush these buffered events, even with SQPOLL enabled.
+        //
+        // Without this, completions remain stuck in kernel's internal buffer
+        // after draining CQ, causing missing completion notifications.
+        let need_syscall_for_overflow = sq_cq_overflow && self.params.is_feature_nodrop();
+
+        if want > 0 || self.params.is_setup_iopoll() || sq_cq_overflow {
             flags |= sys::IORING_ENTER_GETEVENTS;
         }
 
         if self.params.is_setup_sqpoll() {
+            // See discussion in [`SubmissionQueue::need_wakeup`].
+            atomic::fence(atomic::Ordering::SeqCst);
             if self.sq_need_wakeup() {
                 flags |= sys::IORING_ENTER_SQ_WAKEUP;
-            } else if want == 0 {
+            } else if want == 0 && !need_syscall_for_overflow {
                 // The kernel thread is polling and hasn't fallen asleep, so we don't need to tell
                 // it to process events or wake it up
+
+                // However, if the CQ ring is overflown, we need to tell the kernel to process events
+                // by calling io_uring_enter with the IORING_ENTER_GETEVENTS flag.
                 return Ok(len);
             }
         }
@@ -137,6 +156,11 @@ impl<'a> Submitter<'a> {
         unsafe { self.enter::<libc::sigset_t>(len as _, want as _, flags, None) }
     }
 
+    /// Submit all queued submission queue events to the kernel and wait for at least `want`
+    /// completion events to complete with additional options
+    ///
+    /// You can specify a set of signals to mask and a timeout for operation, see
+    /// [`SubmitArgs`](types::SubmitArgs) for more details
     pub fn submit_with_args(
         &self,
         want: usize,
@@ -145,14 +169,19 @@ impl<'a> Submitter<'a> {
         let len = self.sq_len();
         let mut flags = sys::IORING_ENTER_EXT_ARG;
 
-        if want > 0 || self.params.is_setup_iopoll() || self.sq_cq_overflow() {
+        let sq_cq_overflow = self.sq_cq_overflow();
+        let need_syscall = sq_cq_overflow & self.params.is_feature_nodrop();
+
+        if want > 0 || self.params.is_setup_iopoll() || sq_cq_overflow {
             flags |= sys::IORING_ENTER_GETEVENTS;
         }
 
         if self.params.is_setup_sqpoll() {
+            // See discussion in [`SubmissionQueue::need_wakeup`].
+            atomic::fence(atomic::Ordering::SeqCst);
             if self.sq_need_wakeup() {
                 flags |= sys::IORING_ENTER_SQ_WAKEUP;
-            } else if want == 0 {
+            } else if want == 0 && !need_syscall {
                 // The kernel thread is polling and hasn't fallen asleep, so we don't need to tell
                 // it to process events or wake it up
                 return Ok(len);
@@ -186,6 +215,108 @@ impl<'a> Submitter<'a> {
         .map(drop)
     }
 
+    /// Update a range of fixed buffers starting at `offset`.
+    ///
+    /// This is required to use buffers registered using
+    /// [`register_buffers_sparse`](Self::register_buffers_sparse),
+    /// although it can be also be used with [`register_buffers`](Self::register_buffers).
+    ///
+    /// See [`register_buffers2`](Self::register_buffers2)
+    /// for more information about resource tagging.
+    ///
+    /// Available since Linux 5.13.
+    ///
+    /// # Safety
+    ///
+    /// Developers must ensure that the `iov_base` and `iov_len` values are valid and will
+    /// be valid until buffers are unregistered or the ring destroyed, otherwise undefined
+    /// behaviour may occur.
+    pub unsafe fn register_buffers_update(
+        &self,
+        offset: u32,
+        bufs: &[libc::iovec],
+        tags: Option<&[u64]>,
+    ) -> io::Result<()> {
+        let nr = tags
+            .as_ref()
+            .map_or(bufs.len(), |tags| bufs.len().min(tags.len()));
+
+        let rr = sys::io_uring_rsrc_update2 {
+            nr: nr as _,
+            data: bufs.as_ptr() as _,
+            tags: tags.map(|tags| tags.as_ptr() as _).unwrap_or(0),
+            offset,
+            ..Default::default()
+        };
+
+        execute(
+            self.fd.as_raw_fd(),
+            sys::IORING_REGISTER_BUFFERS_UPDATE,
+            cast_ptr::<sys::io_uring_rsrc_update2>(&rr).cast(),
+            std::mem::size_of::<sys::io_uring_rsrc_update2>() as _,
+        )
+        .map(drop)
+    }
+
+    /// Variant of [`register_buffers`](Self::register_buffers)
+    /// with resource tagging.
+    ///
+    /// `tags` should be the same length as `bufs` and contain the
+    /// tag value corresponding to the buffer at the same index.
+    ///
+    /// If a tag is zero, then tagging for this particular resource
+    /// (a buffer in this case) is disabled. Otherwise, after the
+    /// resource had been unregistered and it's not used anymore,
+    /// a CQE will be posted with `user_data` set to the specified
+    /// tag and all other fields zeroed.
+    ///
+    /// Available since Linux 5.13.
+    ///
+    /// # Safety
+    ///
+    /// Developers must ensure that the `iov_base` and `iov_len` values are valid and will
+    /// be valid until buffers are unregistered or the ring destroyed, otherwise undefined
+    /// behaviour may occur.
+    pub unsafe fn register_buffers2(&self, bufs: &[libc::iovec], tags: &[u64]) -> io::Result<()> {
+        let rr = sys::io_uring_rsrc_register {
+            nr: bufs.len().min(tags.len()) as _,
+            data: bufs.as_ptr() as _,
+            tags: tags.as_ptr() as _,
+            ..Default::default()
+        };
+        execute(
+            self.fd.as_raw_fd(),
+            sys::IORING_REGISTER_BUFFERS2,
+            cast_ptr::<sys::io_uring_rsrc_register>(&rr).cast(),
+            std::mem::size_of::<sys::io_uring_rsrc_register>() as _,
+        )
+        .map(drop)
+    }
+
+    /// Registers an empty table of nr fixed buffers buffers.
+    ///
+    /// These must be updated before use, using eg.
+    /// [`register_buffers_update`](Self::register_buffers_update).
+    ///
+    /// See [`register_buffers`](Self::register_buffers)
+    /// for more information about fixed buffers.
+    ///
+    /// Available since Linux 5.13.
+    pub fn register_buffers_sparse(&self, nr: u32) -> io::Result<()> {
+        let rr = sys::io_uring_rsrc_register {
+            nr,
+            flags: sys::IORING_RSRC_REGISTER_SPARSE,
+            ..Default::default()
+        };
+        execute(
+            self.fd.as_raw_fd(),
+            sys::IORING_REGISTER_BUFFERS2,
+            cast_ptr::<sys::io_uring_rsrc_register>(&rr).cast(),
+            std::mem::size_of::<sys::io_uring_rsrc_register>() as _,
+        )
+        .map(drop)
+    }
+
     /// Registers an empty file table of nr_files number of file descriptors. The sparse variant is
     /// available in kernels 5.19 and later.
     ///
@@ -324,6 +455,8 @@ impl<'a> Submitter<'a> {
     ///
     /// You do not need to explicitly call this before dropping the [`IoUring`](crate::IoUring), as
     /// it will be cleaned up by the kernel automatically.
+    ///
+    /// Available since Linux 5.1.
     pub fn unregister_buffers(&self) -> io::Result<()> {
         execute(
             self.fd.as_raw_fd(),
@@ -396,6 +529,30 @@ impl<'a> Submitter<'a> {
         .map(drop)
     }
 
+    /// Tell io_uring on what CPUs the async workers can run. By default, async workers
+    /// created by io_uring will inherit the CPU mask of its parent. This is usually
+    /// all the CPUs in the system, unless the parent is being run with a limited set.
+    pub fn register_iowq_aff(&self, cpu_set: &libc::cpu_set_t) -> io::Result<()> {
+        execute(
+            self.fd.as_raw_fd(),
+            sys::IORING_REGISTER_IOWQ_AFF,
+            cpu_set as *const _ as *const libc::c_void,
+            mem::size_of::<libc::cpu_set_t>() as u32,
+        )
+        .map(drop)
+    }
+
+    /// Undoes a CPU mask previously set with register_iowq_aff
+    pub fn unregister_iowq_aff(&self) -> io::Result<()> {
+        execute(
+            self.fd.as_raw_fd(),
+            sys::IORING_UNREGISTER_IOWQ_AFF,
+            ptr::null(),
+            0,
+        )
+        .map(drop)
+    }
+
     /// Get and/or set the limit for number of io_uring worker threads per NUMA
     /// node. `max[0]` holds the limit for bounded workers, which process I/O
     /// operations expected to be bound in time, that is I/O on regular files or
@@ -427,11 +584,36 @@ impl<'a> Submitter<'a> {
     /// Developers must ensure that the `ring_addr` and its length represented by `ring_entries`
     /// are valid and will be valid until the bgid is unregistered or the ring destroyed,
     /// otherwise undefined behaviour may occur.
+    #[deprecated(note = "please use `register_buf_ring_with_flags` instead")]
     pub unsafe fn register_buf_ring(
         &self,
         ring_addr: u64,
         ring_entries: u16,
         bgid: u16,
+    ) -> io::Result<()> {
+        self.register_buf_ring_with_flags(ring_addr, ring_entries, bgid, 0)
+    }
+
+    /// Register buffer ring for provided buffers.
+    ///
+    /// Details can be found in the io_uring_register_buf_ring.3 man page.
+    ///
+    /// If the register command is not supported, or the ring_entries value exceeds
+    /// 32768, the InvalidInput error is returned.
+    ///
+    /// Available since 5.19.
+    ///
+    /// # Safety
+    ///
+    /// Developers must ensure that the `ring_addr` and its length represented by `ring_entries`
+    /// are valid and will be valid until the bgid is unregistered or the ring destroyed,
+    /// otherwise undefined behaviour may occur.
+    pub unsafe fn register_buf_ring_with_flags(
+        &self,
+        ring_addr: u64,
+        ring_entries: u16,
+        bgid: u16,
+        flags: u16,
     ) -> io::Result<()> {
         // The interface type for ring_entries is u32 but the same interface only allows a u16 for
         // the tail to be specified, so to try and avoid further confusion, we limit the
@@ -441,6 +623,7 @@ impl<'a> Submitter<'a> {
             ring_addr,
             ring_entries: ring_entries as _,
             bgid,
+            flags,
             ..Default::default()
         };
         execute(
@@ -475,7 +658,7 @@ impl<'a> Submitter<'a> {
     /// except that it completes synchronously.
     ///
     /// Cancellation can target a specific request, or all requests matching some criteria. The
-    /// [CancelBuilder](types::CancelBuilder) builder supports describing the match criteria for cancellation.
+    /// [`CancelBuilder`] builder supports describing the match criteria for cancellation.
     ///
     /// An optional `timeout` can be provided to specify how long to wait for matched requests to be
     /// canceled. If no timeout is provided, the default is to wait indefinitely.
@@ -506,22 +689,17 @@ impl<'a> Submitter<'a> {
             tv_nsec: -1,
         });
         let user_data = builder.user_data.unwrap_or(0);
-        let fd = builder
-            .fd
-            .map(|target| match target {
-                types::sealed::Target::Fd(fd) => fd,
-                types::sealed::Target::Fixed(idx) => idx as i32,
-            })
-            .unwrap_or(-1);
         let flags = builder.flags.bits();
+        let fd = builder.to_fd();
 
         let arg = sys::io_uring_sync_cancel_reg {
             addr: user_data,
             fd,
             flags,
             timeout: timespec,
-            pad: [0u64; 4],
+            ..Default::default()
         };
+
         execute(
             self.fd.as_raw_fd(),
             sys::IORING_REGISTER_SYNC_CANCEL,
@@ -530,4 +708,17 @@ impl<'a> Submitter<'a> {
         )
         .map(drop)
     }
+
+    /// Register a netdev hw rx queue for zerocopy.
+    ///
+    /// Available since 6.15.
+    pub fn register_ifq(&self, reg: &sys::io_uring_zcrx_ifq_reg) -> io::Result<()> {
+        execute(
+            self.fd.as_raw_fd(),
+            sys::IORING_REGISTER_ZCRX_IFQ,
+            cast_ptr::<sys::io_uring_zcrx_ifq_reg>(reg) as _,
+            1,
+        )
+        .map(drop)
+    }
 }
