@@ -49,6 +49,16 @@ use std::num::NonZeroU32;
 use std::os::unix::io::RawFd;
 
 pub use sys::__kernel_rwf_t as RwFlags;
+pub use sys::{
+    io_uring_region_desc, io_uring_zcrx_area_reg, io_uring_zcrx_cqe, io_uring_zcrx_ifq_reg,
+    io_uring_zcrx_rqe, IORING_MEM_REGION_TYPE_USER, IORING_ZCRX_AREA_SHIFT, IOU_PBUF_RING_INC,
+    IOU_PBUF_RING_MMAP,
+};
+
+// From linux/io_uring.h
+//
+// NOTE: bindgen skips this due to the expression so we define it manually.
+pub const IORING_ZCRX_AREA_MASK: u64 = !((1u64 << IORING_ZCRX_AREA_SHIFT) - 1);
 
 /// Opaque types, you should use [`statx`](struct@libc::statx) instead.
 #[repr(C)]
@@ -80,14 +90,15 @@ bitflags! {
     /// Options for [`Timeout`](super::Timeout).
     ///
     /// The default behavior is to treat the timespec as a relative time interval. `flags` may
-    /// contain [`types::TimeoutFlags::ABS`] to indicate the timespec represents an absolute
+    /// contain [`TimeoutFlags::ABS`] to indicate the timespec represents an absolute
     /// time. When an absolute time is being specified, the kernel will use its monotonic clock
     /// unless one of the following flags is set (they may not both be set):
-    /// [`types::TimeoutFlags::BOOTTIME`] or [`types::TimeoutFlags::REALTIME`].
+    /// [`TimeoutFlags::BOOTTIME`] or [`TimeoutFlags::REALTIME`].
     ///
-    /// The default behavior when the timeout expires is to return a CQE with -libc::ETIME in
-    /// the res field. To change this behavior to have zero returned, include
-    /// [`types::TimeoutFlags::ETIME_SUCCESS`].
+    /// The default behavior when the timeout expires is to sever dependent links, as a failed
+    /// request normally would. To keep the links untouched include [`TimeoutFlags::ETIME_SUCCESS`].
+    /// CQE will still contain -libc::ETIME in the res field
+    #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
     pub struct TimeoutFlags: u32 {
         const ABS = sys::IORING_TIMEOUT_ABS;
 
@@ -98,11 +109,14 @@ bitflags! {
         const LINK_TIMEOUT_UPDATE = sys::IORING_LINK_TIMEOUT_UPDATE;
 
         const ETIME_SUCCESS = sys::IORING_TIMEOUT_ETIME_SUCCESS;
+
+        const MULTISHOT = sys::IORING_TIMEOUT_MULTISHOT;
     }
 }
 
 bitflags! {
     /// Options for [`Fsync`](super::Fsync).
+    #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
     pub struct FsyncFlags: u32 {
         const DATASYNC = sys::IORING_FSYNC_DATASYNC;
     }
@@ -111,6 +125,7 @@ bitflags! {
 bitflags! {
     /// Options for [`AsyncCancel`](super::AsyncCancel) and
     /// [`Submitter::register_sync_cancel`](super::Submitter::register_sync_cancel).
+    #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
     pub(crate) struct AsyncCancelFlags: u32 {
         /// Cancel all requests that match the given criteria, rather
         /// than just canceling the first one found.
@@ -236,7 +251,7 @@ impl<'prev, 'now> SubmitArgs<'prev, 'now> {
         let args = sys::io_uring_getevents_arg {
             sigmask: 0,
             sigmask_sz: 0,
-            pad: 0,
+            min_wait_usec: 0,
             ts: 0,
         };
 
@@ -248,6 +263,9 @@ impl<'prev, 'now> SubmitArgs<'prev, 'now> {
     }
 
     #[inline]
+    /// Signals to mask during waiting for the result
+    ///
+    /// Masked signals will be restored after submit operation returns
     pub fn sigmask<'new>(mut self, sigmask: &'new libc::sigset_t) -> SubmitArgs<'now, 'new> {
         self.args.sigmask = cast_ptr(sigmask) as _;
         self.args.sigmask_sz = std::mem::size_of::<libc::sigset_t>() as _;
@@ -260,6 +278,7 @@ impl<'prev, 'now> SubmitArgs<'prev, 'now> {
     }
 
     #[inline]
+    /// Timeout for submit operation
     pub fn timespec<'new>(mut self, timespec: &'new Timespec) -> SubmitArgs<'now, 'new> {
         self.args.ts = cast_ptr(timespec) as _;
 
@@ -320,7 +339,7 @@ impl BufRingEntry {
     /// of the buf_ring for the resv field to even be considered the tail field of the ring.
     /// The entry must also be properly initialized.
     pub unsafe fn tail(ring_base: *const BufRingEntry) -> *const u16 {
-        &(*ring_base).0.resv
+        std::ptr::addr_of!((*ring_base).0.resv)
     }
 }
 
@@ -393,6 +412,7 @@ impl<'buf> RecvMsgOut<'buf> {
     /// is the same content provided as input to the corresponding SQE
     /// (only `msg_namelen` and `msg_controllen` fields are relevant).
     #[allow(clippy::result_unit_err)]
+    #[allow(clippy::useless_conversion)]
     pub fn parse(buffer: &'buf [u8], msghdr: &libc::msghdr) -> Result<Self, ()> {
         let msghdr_name_len = usize::try_from(msghdr.msg_namelen).unwrap();
         let msghdr_control_len = usize::try_from(msghdr.msg_controllen).unwrap();
